import logging

l = logging.getLogger("archinfo.arch_x86")

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

from .arch import Arch, register_arch, Endness, Register
from .tls import TLSArchInfo
from .archerror import ArchError

class ArchX86(Arch):
    def __init__(self, endness=Endness.LE):
        if endness != Endness.LE:
            raise ArchError('Arch i386 must be little endian')
        super(ArchX86, self).__init__(endness)
        if self.vex_archinfo:
            self.vex_archinfo['x86_cr0'] = 0xFFFFFFFF

    @property
    def capstone_x86_syntax(self):
        """
        Get the current syntax Capstone uses for x86. It can be 'intel' or 'at&t'

        :return: Capstone's current x86 syntax
        :rtype: str
        """

        return self._cs_x86_syntax

    @capstone_x86_syntax.setter
    def capstone_x86_syntax(self, new_syntax):
        """
        Set the syntax that Capstone outputs for x86.
        """

        if new_syntax not in ('intel', 'at&t'):
            raise ArchError('Unsupported Capstone x86 syntax. It must be either "intel" or "at&t".')

        if new_syntax != self._cs_x86_syntax:
            self._cs = None
            self._cs_x86_syntax = new_syntax

    def _configure_capstone(self):
        self._cs.syntax = _capstone.CS_OPT_SYNTAX_ATT if self._cs_x86_syntax == 'at&t' else _capstone.CS_OPT_SYNTAX_INTEL

    @property
    def keystone_x86_syntax(self):
        """
        Get the current syntax Keystone uses for x86. It can be 'intel',
        'at&t', 'nasm', 'masm', 'gas' or 'radix16'

        :return: Keystone's current x86 syntax
        :rtype: str
        """

        return self._ks_x86_syntax

    @keystone_x86_syntax.setter
    def keystone_x86_syntax(self, new_syntax):
        """
        Set the syntax that Keystone uses for x86.
        """

        if new_syntax not in ('intel', 'at&t', 'nasm', 'masm', 'gas', 'radix16'):
            raise ArchError('Unsupported Keystone x86 syntax. It must be one of the following: '
                            '"intel", "at&t", "nasm", "masm", "gas" or "radix16".')

        if new_syntax != self._ks_x86_syntax:
            self._ks = None
            self._ks_x86_syntax = new_syntax

    def _configure_keystone(self):
        if self._ks_x86_syntax == 'at&t':
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_ATT
        elif self._ks_x86_syntax == 'nasm':
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_NASM
        elif self._ks_x86_syntax == 'masm':
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_MASM
        elif self._ks_x86_syntax == 'gas':
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_GAS
        elif self._ks_x86_syntax == 'radix16':
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_RADIX16
        else:
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_INTEL

    bits = 32
    vex_arch = "VexArchX86"
    name = "X86"
    qemu_name = 'i386'
    ida_processor = 'metapc'
    linux_name = 'i386'
    triplet = 'i386-linux-gnu'
    max_inst_bytes = 15
    call_sp_fix = -8
    ret_offset = 8
    vex_conditional_helpers = True
    syscall_num_offset = 8
    call_pushes_ret = True
    stack_change = -4
    memory_endness = Endness.LE
    register_endness = Endness.LE
    sizeof = {'short': 16, 'int': 32, 'long': 32, 'long long': 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_X86
        cs_mode = _capstone.CS_MODE_32 + _capstone.CS_MODE_LITTLE_ENDIAN
    _cs_x86_syntax = None # Set it to 'att' in order to use AT&T syntax for x86
    if _keystone:
        ks_arch = _keystone.KS_ARCH_X86
        ks_mode = _keystone.KS_MODE_32 + _keystone.KS_MODE_LITTLE_ENDIAN
    _ks_x86_syntax = None
    uc_arch = _unicorn.UC_ARCH_X86 if _unicorn else None
    uc_mode = (_unicorn.UC_MODE_32 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    uc_const = _unicorn.x86_const if _unicorn else None
    uc_prefix = "UC_X86_" if _unicorn else None
    function_prologs = [
        br"\x8b\xff\x55\x8b\xec", # mov edi, edi; push ebp; mov ebp, esp
        br"\x55\x8b\xec", # push ebp; mov ebp, esp
        br"\x55\x89\xe5",  # push ebp; mov ebp, esp
        br"\x55\x57\x56",  # push ebp; push edi; push esi
        # mov eax, 0x000000??; (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) sub esp
        br"\xb8[\x00-\xff]\x00\x00\x00[\x50\x51\x52\x53\x55\x56\x57]{0,7}\x8b[\x00-\xff]{2}",
        # (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) sub esp
        br"[\x50\x51\x52\x53\x55\x56\x57]{1,7}\x83\xec[\x00-\xff]{2,4}",
        # (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) mov xxx, xxx
        br"[\x50\x51\x52\x53\x55\x56\x57]{1,7}\x8b[\x00-\xff]{2}",
        br"(\x81|\x83)\xec",  # sub xxx %esp
    ]
    function_epilogs = {
        br"\xc9\xc3", # leave; ret
        br"([^\x41][\x50-\x5f]{1}|\x41[\x50-\x5f])\xc3", # pop <reg>; ret
        br"[^\x48][\x83,\x81]\xc4([\x00-\xff]{1}|[\x00-\xff]{4})\xc3", #  add esp, <siz>; retq
    }
    ret_instruction = b"\xc3"
    nop_instruction = b"\x90"
    instruction_alignment = 1
    register_list = [
        Register(name='eax', size=4, subregisters=[('ax', 0, 2),
                                                   ('al', 0, 1),
                                                   ('ah', 1, 1)],
                 general_purpose=True, argument=True, linux_entry_value=0x1C),
        Register(name='ecx', size=4, subregisters=[('cx', 0, 2),
                                                   ('cl', 0, 1),
                                                   ('ch', 1, 1)],
                 general_purpose=True, argument=True),
        Register(name='edx', size=4, subregisters=[('dx', 0, 2),
                                                   ('dl', 0, 1),
                                                   ('dh', 1, 1)],
                 general_purpose=True, argument=True, linux_entry_value='ld_destructor'),
        Register(name='ebx', size=4, subregisters=[('bx', 0, 2),
                                                   ('bl', 0, 1),
                                                   ('bh', 1, 1)],
                 general_purpose=True, argument=True),
        Register(name='esp', size=4, alias_names=('sp',),
                 general_purpose=True, default_value=(Arch.initial_sp, True, 'global')),
        Register(name='ebp', size=4, alias_names=('bp',),
                 general_purpose=True, argument=True, linux_entry_value=0),
        Register(name='esi', size=4, subregisters=[('si', 0, 2),
                                                   ('sil', 0, 1),
                                                   ('sih', 1, 1)],
                 general_purpose=True, argument=True),
        Register(name='edi', size=4, subregisters=[('di', 0, 2),
                                                   ('dil', 0, 1),
                                                   ('dih', 1, 1)],
                 general_purpose=True, argument=True),
        Register(name='cc_op', size=4, default_value=(0, False, None), concrete=False, artificial=True),
        Register(name='cc_dep1', size=4, concrete=False, artificial=True),
        Register(name='cc_dep2', size=4, concrete=False, artificial=True),
        Register(name='cc_ndep', size=4, concrete=False, artificial=True),
        Register(name='d', size=4, alias_names=('dflag',),
                 default_value=(1, False, None), concrete=False),
        Register(name='id', size=4, alias_names=('idflag',),
                 default_value=(1, False, None), concrete=False),
        Register(name='ac', size=4, alias_names=('acflag',),
                 default_value=(0, False, None), concrete=False),
        Register(name='eip', size=4, alias_names=('ip', 'pc')),
        Register(name='fpreg', size=64, subregisters=[('mm0', 0, 8),
                                                      ('mm1', 8, 8),
                                                      ('mm2', 16, 8),
                                                      ('mm3', 24, 8),
                                                      ('mm4', 32, 8),
                                                      ('mm5', 40, 8),
                                                      ('mm6', 48, 8),
                                                      ('mm7', 56, 8)],
                 alias_names=('fpu_regs',), floating_point=True, concrete=False),
        Register(name='fptag', size=8, alias_names=('fpu_tags',),
                 floating_point=True, default_value=(0, False, None)),
        Register(name='fpround', size=4, floating_point=True,
                 default_value=(0, False, None)),
        Register(name='fc3210', size=4, floating_point=True),
        Register(name='ftop', size=4, floating_point=True,
                 default_value=(0, False, None)),
        Register(name='sseround', size=4, vector=True,
                 default_value=(0, False, None)),
        Register(name='xmm0', size=16, vector=True),
        Register(name='xmm1', size=16, vector=True),
        Register(name='xmm2', size=16, vector=True),
        Register(name='xmm3', size=16, vector=True),
        Register(name='xmm4', size=16, vector=True),
        Register(name='xmm5', size=16, vector=True),
        Register(name='xmm6', size=16, vector=True),
        Register(name='xmm7', size=16, vector=True),
        Register(name='cs', size=2),
        Register(name='ds', size=2),
        Register(name='es', size=2),
        Register(name='fs', size=2, default_value=(0, False, None), concrete=False),
        Register(name='gs', size=2, default_value=(0, False, None), concrete=False),
        Register(name='ss', size=2),
        Register(name='ldt', size=8, default_value=(0, False, None), concrete=False),
        Register(name='gdt', size=8, default_value=(0, False, None), concrete=False),
        Register(name='emnote', size=4),
        Register(name='cmstart', size=4),
        Register(name='cmlen', size=4),
        Register(name='nraddr', size=4),
        Register(name='sc_class', size=4),
        Register(name='ip_at_syscall', size=4, concrete=False, artificial=True),
    ]

    symbol_type_translation = {
        10: 'STT_GNU_IFUNC',
        'STT_LOOS': 'STT_GNU_IFUNC'
    }
    lib_paths = ['/lib32', '/usr/lib32']
    got_section_name = '.got.plt'
    ld_linux_name = 'ld-linux.so.2'
    elf_tls = TLSArchInfo(2, 56, [8], [4], [0], 0, 0)

register_arch([r'.*i?\d86|.*x32|.*x86|.*metapc'], 32, Endness.LE, ArchX86)




import logging
import re

from .arch import Arch, Endness, register_arch

l = logging.getLogger('archinfo.arch_soot')


class SootMethodDescriptor(object):

    __slots__ = ['class_name', 'name', 'params', '_soot_method', 'ret']

    def __init__(self, class_name, name, params, soot_method=None, ret_type =None):
        self.class_name = class_name
        self.name = name
        self.params = params
        self._soot_method = soot_method
        self.ret = ret_type

    def __repr__(self):
        return "%s.%s(%s)" % (self.class_name, self.name, ", ".join(self.params))

    def __hash__(self):
        return hash((self.class_name, self.name, self.params))

    def __eq__(self, other):
        return isinstance(other, SootMethodDescriptor) and \
            self.class_name == other.class_name and \
            self.name == other.name and \
            self.params == other.params

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.__repr__() < other.__repr__()

    def __gt__(self, other):
        return self.__repr__() > other.__repr__()

    def __le__(self, other):
        return self.__repr__() <= other.__repr__()

    def __ge__(self, other):
        return self.__repr__() >= other.__repr__()

    def address(self, block_idx=0, stmt_idx=0):
        """
        :return Address of the method.
        :rtype: SootAddressDescriptor
        """
        return SootAddressDescriptor(self, block_idx, stmt_idx)

    @property
    def fullname(self):
        """
        :return the full name of the method (class name + method name)
        """
        return "%s.%s" % (self.class_name, self.name)

    @property
    def symbolic(self):
        return False

    @property
    def is_loaded(self):
        """
        :return: True, if the method is loaded in CLE and thus infos about attrs,
                 ret and exceptions are available.
        """
        return self._soot_method is not None

    @property
    def attrs(self):
        return self._soot_method.attrs if self.is_loaded else []

    @property
    def exceptions(self):
        return self._soot_method.exceptions if self.is_loaded else []

    @property
    def block_by_label(self):
        return self._soot_method.block_by_label if self.is_loaded else None

    # @property
    # def ret(self):
    #     return self._soot_method.ret if self.is_loaded else []

    @property
    def addr(self):
        """
        :return: the soot address description of the entry point of the method
        """
        return SootAddressDescriptor(self, 0, 0)


    def matches_with_native_name(self, native_method):
        """
        The name of native methods are getting encoded, s.t. they translate into
        valid C function names. This method indicates if the name of the given
        native method matches the name of the soot method.

        :return: True, if name of soot method matches the mangled native name.
        """

        if "__" in native_method:
            # if native methods are overloaded, two underscores are used
            native_method, params_sig = native_method.split('__')
            params = ArchSoot.decode_parameter_list_signature(params_sig)
            # check function signature
            if params != self.params:
                return False

        # demangle native name
        native_method = native_method.replace('_1', '_')
        # TODO unicode escaping

        method_native_name = "Java_{class_name}_{method_name}".format(
                              class_name=self.class_name.replace('.', '_'),
                              method_name=self.name)

        return native_method == method_native_name

    @classmethod
    def from_string(cls, tstr):
        # this should be the opposite of repr
        tstr = tstr.strip()
        class_and_method, tparams = tstr.split("(")
        params_str = tparams.split(")")[0]
        if params_str == "":
            params = tuple()
        else:
            params = tuple([t.strip() for t in params_str.split(",")])
        class_name, _, method = class_and_method.rpartition(".")
        return cls(class_name, method, params)

    @classmethod
    def from_soot_method(cls, soot_method):
        return cls(class_name=soot_method.class_name,
                   name=soot_method.name,
                   params=soot_method.params,
                   soot_method=soot_method,
                   ret_type=soot_method.ret)


class SootAddressDescriptor(object):

    __slots__ = ['method', 'block_idx', 'stmt_idx']

    def __init__(self, method, block_idx, stmt_idx):

        if not isinstance(method, SootMethodDescriptor):
            raise ValueError('The parameter "method" must be an '
                             'instance of SootMethodDescriptor.')

        self.method = method
        self.block_idx = block_idx
        self.stmt_idx = stmt_idx

    def __repr__(self):
        return "<%r+(%s:%s)>" % (self.method,
                                 self.block_idx,
                                 '%d' % self.stmt_idx if self.stmt_idx is not None else '[0]')

    def __hash__(self):
        return hash((self.method, self.stmt_idx))

    def __eq__(self, other):
        return isinstance(other, SootAddressDescriptor) and \
            self.method == other.method and \
            self.block_idx == other.block_idx and \
            self.stmt_idx == other.stmt_idx

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.__repr__() < other.__repr__()

    def __gt__(self, other):
        return self.__repr__() > other.__repr__()

    def __le__(self, other):
        return self.__repr__() <= other.__repr__()

    def __ge__(self, other):
        return self.__repr__() >= other.__repr__()

    def __add__(self, stmts_offset):
        if not isinstance(stmts_offset, int):
            raise TypeError('The stmts_offset must be an int or a long.')
        s = self.copy()
        s.stmt_idx += stmts_offset
        return s

    def copy(self):
        return SootAddressDescriptor(method=self.method,
                                     block_idx=self.block_idx,
                                     stmt_idx=self.stmt_idx)

    @property
    def symbolic(self):
        return False


class SootAddressTerminator(SootAddressDescriptor):

    __slots__ = []

    def __init__(self):
        dummy_method = SootMethodDescriptor("dummy", "dummy", tuple())
        super(SootAddressTerminator, self).__init__(dummy_method, 0, 0)

    def __repr__(self):
        return "<Terminator>"


class SootFieldDescriptor(object):

    __slots__ = ['class_name', 'name', 'type']

    def __init__(self, class_name, name, type_):
        self.class_name = class_name
        self.name = name
        self.type = type_

    def __repr__(self):
        return "%s.%s" % (self.class_name, self.name)

    def __hash__(self):
        return hash((self.class_name, self.name, self.type))

    def __eq__(self, other):
        return isinstance(other, SootFieldDescriptor) and \
            self.class_name == other.class_name and \
            self.name == other.name and \
            self.type == other.type

    def __ne__(self, other):
        return not self == other


class SootClassDescriptor(object):

    __slots__ = ['name', '_soot_class']

    def __init__(self, name, soot_class=None):
        self.name = name
        self._soot_class = soot_class

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, SootClassDescriptor) and \
            self.name == other.name

    def __ne__(self, other):
        return not self == other

    @property
    def is_loaded(self):
        """
        :return: True, if the class is loaded in CLE and thus info about field,
                 methods, ... are available.
        """
        return self._soot_class is not None

    @property
    def fields(self):
        return self._soot_class.fields if self.is_loaded else None

    @property
    def methods(self):
        return self._soot_class.methods if self.is_loaded else None

    @property
    def superclass_name(self):
        return self._soot_class.super_class if self.is_loaded else None

    @property
    def type(self):
        return "java.lang.Class"


class SootNullConstant(object):

    def __init__(self):
        pass

    def __repr__(self):
        return 'null'

    def __hash__(self):
        return hash('null')

    def __eq__(self, other):
        return isinstance(other, SootNullConstant)

    def __ne__(self, other):
        return not self == other


class SootArgument(object):
    """
    Typed Java argument.
    """

    __slots__ = ['value', 'type', 'is_this_ref']

    def __init__(self, value, type_, is_this_ref=False):
        """
        :param value:        Value of the argument
        :param type_:        Type of the argument
        :param is_this_ref:  Indicates whether the argument represents the
                             'this' reference, i.e. the object on which an
                             instance method is invoked.
        """
        self.value = value
        self.type = type_
        self.is_this_ref = is_this_ref

    def __repr__(self):
        return "%s (%s)" % (self.value, self.type)


class ArchSoot(Arch):
    def __init__(self, endness=Endness.LE):
        super(ArchSoot, self).__init__(endness)

    name = 'Soot'

    vex_arch = None  # No VEX support
    qemu_name = None  # No Qemu/Unicorn-engine support
    bits = 64
    address_types = (SootAddressDescriptor, )
    function_address_types = (SootMethodDescriptor, )

    # Size of native counterparts of primitive Java types
    sizeof = {'boolean':  8,
              'byte':  8,
              'char': 16,
              'short': 16,
              'int': 32,
              'long': 64,
              'float': 32,
              'double': 64}

    primitive_types = ['boolean',
                       'byte',
                       'char',
                       'short',
                       'int',
                       'long',
                       'float',
                       'double']

    sig_dict = {'Z': 'boolean',
                'B': 'byte',
                'C': 'char',
                'S': 'short',
                'I': 'int',
                'J': 'long',
                'F': 'float',
                'D': 'double',
                'V': 'void'}

    @staticmethod
    def decode_type_signature(type_sig):
        if not type_sig:
            return None
        # try to translate signature as a primitive type
        if type_sig in ArchSoot.sig_dict:
            return ArchSoot.sig_dict[type_sig]
        # java classes are encoded as 'Lclass_name;'
        if type_sig.startswith('L') and type_sig.endswith(';'):
            return type_sig[1:-1]
        raise ValueError(type_sig)

    @staticmethod
    def decode_parameter_list_signature(param_sig):
        return tuple(ArchSoot.decode_type_signature(param)
                     for param in re.findall(r'([\[]*[ZBCSIJFDV]|[\[]*L.+;)', param_sig))

    @staticmethod
    def decode_method_signature(method_sig):
        # signature format follows the pattern: (param_sig)ret_sig
        match = re.search(r'\((.*)\)(.*)', method_sig)
        param_sig, ret_sig = match.group(1), match.group(2)
        # decode types
        params_types = ArchSoot.decode_parameter_list_signature(param_sig)
        ret_type = ArchSoot.decode_type_signature(ret_sig)
        l.debug("Decoded method signature '%s' as params=%s and ret=%s",
                method_sig, params_types, ret_type)
        return params_types, ret_type

    def library_search_path(self, pedantic=False):
        """
        Since Java is mostly system independent, we cannot return system
        specific paths.

        :return: empty list
        """
        return []


register_arch(['soot'], 8, Endness.LE, ArchSoot)



import logging

l = logging.getLogger("archinfo.arch_aarch64")

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

from .arch import Arch, register_arch, Endness, Register
from .tls import TLSArchInfo

class ArchAArch64(Arch):
    def __init__(self, endness=Endness.LE):
        super(ArchAArch64, self).__init__(endness)
        if endness == Endness.BE:
            self.ida_processor = 'armb'
            self.function_prologs = set((
                # TODO
            ))
            self.function_epilogs = set((
                # TODO
            ))

    bits = 64
    vex_arch = "VexArchARM64"
    name = "AARCH64"
    qemu_name = 'aarch64'
    ida_processor = 'arm'
    linux_name = 'aarch64'
    triplet = 'aarch64-linux-gnueabihf'
    max_inst_bytes = 4
    ret_offset = 16
    vex_conditional_helpers = True
    syscall_num_offset = 80
    call_pushes_ret = False
    stack_change = -8
    memory_endness = Endness.LE
    register_endness = Endness.LE
    instruction_endness = Endness.LE
    sizeof = {'short': 16, 'int': 32, 'long': 64, 'long long': 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_ARM64
        cs_mode = _capstone.CS_MODE_LITTLE_ENDIAN
    if _keystone:
        ks_arch = _keystone.KS_ARCH_ARM64
        ks_mode = _keystone.KS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_ARM64 if _unicorn else None
    uc_mode = _unicorn.UC_MODE_LITTLE_ENDIAN if _unicorn else None
    uc_const = _unicorn.arm64_const if _unicorn else None
    uc_prefix = "UC_ARM64_" if _unicorn else None
    initial_sp = 0x7ffffffffff0000

    ret_instruction = b"\xC0\x03\x5F\xD6"    # ret
    nop_instruction = b"\x1F\x20\x03\xD5"    # nop
    function_prologs = set((
        #br"\xFD\x7B\xBE\xA9\xFD\x03\x00\x91"
        # TODO
    ))
    function_epilogs = set((
        # TODO
    ))
    instruction_alignment = 4
    register_list = [
        Register(name='x0', size=8, subregisters=[('w0', 0, 4)], alias_names=('r0',),
                 general_purpose=True, argument=True, linux_entry_value='ld_destructor'),
        Register(name='x1', size=8, subregisters=[('w1', 0, 4)], alias_names=('r1',),
                 general_purpose=True, argument=True),
        Register(name='x2', size=8, subregisters=[('w2', 0, 4)], alias_names=('r2',),
                 general_purpose=True, argument=True),
        Register(name='x3', size=8, subregisters=[('w3', 0, 4)], alias_names=('r3',),
                 general_purpose=True, argument=True),
        Register(name='x4', size=8, subregisters=[('w4', 0, 4)], alias_names=('r4',),
                 general_purpose=True, argument=True),
        Register(name='x5', size=8, subregisters=[('w5', 0, 4)], alias_names=('r5',),
                 general_purpose=True, argument=True),
        Register(name='x6', size=8, subregisters=[('w6', 0, 4)], alias_names=('r6',),
                 general_purpose=True, argument=True),
        Register(name='x7', size=8, subregisters=[('w7', 0, 4)], alias_names=('r7',),
                 general_purpose=True),
        Register(name='x8', size=8, subregisters=[('w8', 0, 4)], alias_names=('r8',),
                 general_purpose=True),
        Register(name='x9', size=8, subregisters=[('w9', 0, 4)], alias_names=('r9',),
                 general_purpose=True),
        Register(name='x10', size=8, subregisters=[('w10', 0, 4)], alias_names=('r10',),
                 general_purpose=True),
        Register(name='x11', size=8, subregisters=[('w11', 0, 4)], alias_names=('r11',),
                 general_purpose=True),
        Register(name='x12', size=8, subregisters=[('w12', 0, 4)], alias_names=('r12',),
                 general_purpose=True),
        Register(name='x13', size=8, subregisters=[('w13', 0, 4)], alias_names=('r13',),
                 general_purpose=True),
        Register(name='x14', size=8, subregisters=[('w14', 0, 4)], alias_names=('r14',),
                 general_purpose=True),
        Register(name='x15', size=8, subregisters=[('w15', 0, 4)], alias_names=('r15',),
                 general_purpose=True),
        Register(name='x16', size=8, subregisters=[('w16', 0, 4)], alias_names=('r16', 'ip0'),
                 general_purpose=True),
        Register(name='x17', size=8, subregisters=[('w17', 0, 4)], alias_names=('r17', 'ip1'),
                 general_purpose=True),
        Register(name='x18', size=8, subregisters=[('w18', 0, 4)], alias_names=('r18',),
                 general_purpose=True),
        Register(name='x19', size=8, subregisters=[('w19', 0, 4)], alias_names=('r19',),
                 general_purpose=True),
        Register(name='x20', size=8, subregisters=[('w20', 0, 4)], alias_names=('r20',),
                 general_purpose=True),
        Register(name='x21', size=8, subregisters=[('w21', 0, 4)], alias_names=('r21',),
                 general_purpose=True),
        Register(name='x22', size=8, subregisters=[('w22', 0, 4)], alias_names=('r22',),
                 general_purpose=True),
        Register(name='x23', size=8, subregisters=[('w23', 0, 4)], alias_names=('r23',),
                 general_purpose=True),
        Register(name='x24', size=8, subregisters=[('w24', 0, 4)], alias_names=('r24',),
                 general_purpose=True),
        Register(name='x25', size=8, subregisters=[('w25', 0, 4)], alias_names=('r25',),
                 general_purpose=True),
        Register(name='x26', size=8, subregisters=[('w26', 0, 4)], alias_names=('r26',),
                 general_purpose=True),
        Register(name='x27', size=8, subregisters=[('w27', 0, 4)], alias_names=('r27',),
                 general_purpose=True),
        Register(name='x28', size=8, subregisters=[('w28', 0, 4)], alias_names=('r28',),
                 general_purpose=True),
        Register(name='x29', size=8, subregisters=[('w29', 0, 4)], alias_names=('r29', 'fp', 'bp'),
                 general_purpose=True),
        Register(name='x30', size=8, subregisters=[('w30', 0, 4)], alias_names=('r30', 'lr'),
                 general_purpose=True),
        Register(name='xsp', size=8, subregisters=[('wsp', 0, 4)], alias_names=('sp',),
                 general_purpose=True, default_value=(initial_sp, True, 'global')),
        Register(name='pc', size=8, alias_names=('ip',)),
        Register(name='cc_op', size=8, artificial=True),
        Register(name='cc_dep1', size=8, artificial=True),
        Register(name='cc_dep2', size=8, artificial=True),
        Register(name='cc_ndep', size=8, artificial=True),
        Register(name='tpidr_el0', size=8),
        Register(name='q0', size=16, subregisters=[('d0', 0, 8),
                                                   ('s0', 0, 4),
                                                   ('h0', 0, 2),
                                                   ('b0', 0, 1)],
                 alias_names=('v0',), floating_point=True, vector=True),
        Register(name='q1', size=16, subregisters=[('d1', 0, 8),
                                                   ('s1', 0, 4),
                                                   ('h1', 0, 2),
                                                   ('b1', 0, 1)],
                 alias_names=('v1',), floating_point=True, vector=True),
        Register(name='q2', size=16, subregisters=[('d2', 0, 8),
                                                   ('s2', 0, 4),
                                                   ('h2', 0, 2),
                                                   ('b2', 0, 1)],
                 alias_names=('v2',), floating_point=True, vector=True),
        Register(name='q3', size=16, subregisters=[('d3', 0, 8),
                                                   ('s3', 0, 4),
                                                   ('h3', 0, 2),
                                                   ('b3', 0, 1)],
                 alias_names=('v3',), floating_point=True, vector=True),
        Register(name='q4', size=16, subregisters=[('d4', 0, 8),
                                                   ('s4', 0, 4),
                                                   ('h4', 0, 2),
                                                   ('b4', 0, 1)],
                 alias_names=('v4',), floating_point=True, vector=True),
        Register(name='q5', size=16, subregisters=[('d5', 0, 8),
                                                   ('s5', 0, 4),
                                                   ('h5', 0, 2),
                                                   ('b5', 0, 1)],
                 alias_names=('v5',), floating_point=True, vector=True),
        Register(name='q6', size=16, subregisters=[('d6', 0, 8),
                                                   ('s6', 0, 4),
                                                   ('h6', 0, 2),
                                                   ('b6', 0, 1)],
                 alias_names=('v6',), floating_point=True, vector=True),
        Register(name='q7', size=16, subregisters=[('d7', 0, 8),
                                                   ('s7', 0, 4),
                                                   ('h7', 0, 2),
                                                   ('b7', 0, 1)],
                 alias_names=('v7',), floating_point=True, vector=True),
        Register(name='q8', size=16, subregisters=[('d8', 0, 8),
                                                   ('s8', 0, 4),
                                                   ('h8', 0, 2),
                                                   ('b8', 0, 1)],
                 alias_names=('v8',), floating_point=True, vector=True),
        Register(name='q9', size=16, subregisters=[('d9', 0, 8),
                                                   ('s9', 0, 4),
                                                   ('h9', 0, 2),
                                                   ('b9', 0, 1)],
                 alias_names=('v9',), floating_point=True, vector=True),
        Register(name='q10', size=16, subregisters=[('d10', 0, 8),
                                                    ('s10', 0, 4),
                                                    ('h10', 0, 2),
                                                    ('b10', 0, 1)],
                 alias_names=('v10',), floating_point=True, vector=True),
        Register(name='q11', size=16, subregisters=[('d11', 0, 8),
                                                    ('s11', 0, 4),
                                                    ('h11', 0, 2),
                                                    ('b11', 0, 1)],
                 alias_names=('v11',), floating_point=True, vector=True),
        Register(name='q12', size=16, subregisters=[('d12', 0, 8),
                                                    ('s12', 0, 4),
                                                    ('h12', 0, 2),
                                                    ('b12', 0, 1)],
                 alias_names=('v12',), floating_point=True, vector=True),
        Register(name='q13', size=16, subregisters=[('d13', 0, 8),
                                                    ('s13', 0, 4),
                                                    ('h13', 0, 2),
                                                    ('b13', 0, 1)],
                 alias_names=('v13',), floating_point=True, vector=True),
        Register(name='q14', size=16, subregisters=[('d14', 0, 8),
                                                    ('s14', 0, 4),
                                                    ('h14', 0, 2),
                                                    ('b14', 0, 1)],
                 alias_names=('v14',), floating_point=True, vector=True),
        Register(name='q15', size=16, subregisters=[('d15', 0, 8),
                                                    ('s15', 0, 4),
                                                    ('h15', 0, 2),
                                                    ('b15', 0, 1)],
                 alias_names=('v15',), floating_point=True, vector=True),
        Register(name='q16', size=16, subregisters=[('d16', 0, 8),
                                                    ('s16', 0, 4),
                                                    ('h16', 0, 2),
                                                    ('b16', 0, 1)],
                 alias_names=('v16',), floating_point=True, vector=True),
        Register(name='q17', size=16, subregisters=[('d17', 0, 8),
                                                    ('s17', 0, 4),
                                                    ('h17', 0, 2),
                                                    ('b17', 0, 1)],
                 alias_names=('v17',), floating_point=True, vector=True),
        Register(name='q18', size=16, subregisters=[('d18', 0, 8),
                                                    ('s18', 0, 4),
                                                    ('h18', 0, 2),
                                                    ('b18', 0, 1)],
                 alias_names=('v18',), floating_point=True, vector=True),
        Register(name='q19', size=16, subregisters=[('d19', 0, 8),
                                                    ('s19', 0, 4),
                                                    ('h19', 0, 2),
                                                    ('b19', 0, 1)],
                 alias_names=('v19',), floating_point=True, vector=True),
        Register(name='q20', size=16, subregisters=[('d20', 0, 8),
                                                    ('s20', 0, 4),
                                                    ('h20', 0, 2),
                                                    ('b20', 0, 1)],
                 alias_names=('v20',), floating_point=True, vector=True),
        Register(name='q21', size=16, subregisters=[('d21', 0, 8),
                                                    ('s21', 0, 4),
                                                    ('h21', 0, 2),
                                                    ('b21', 0, 1)],
                 alias_names=('v21',), floating_point=True, vector=True),
        Register(name='q22', size=16, subregisters=[('d22', 0, 8),
                                                    ('s22', 0, 4),
                                                    ('h22', 0, 2),
                                                    ('b22', 0, 1)],
                 alias_names=('v22',), floating_point=True, vector=True),
        Register(name='q23', size=16, subregisters=[('d23', 0, 8),
                                                    ('s23', 0, 4),
                                                    ('h23', 0, 2),
                                                    ('b23', 0, 1)],
                 alias_names=('v23',), floating_point=True, vector=True),
        Register(name='q24', size=16, subregisters=[('d24', 0, 8),
                                                    ('s24', 0, 4),
                                                    ('h24', 0, 2),
                                                    ('b24', 0, 1)],
                 alias_names=('v24',), floating_point=True, vector=True),
        Register(name='q25', size=16, subregisters=[('d25', 0, 8),
                                                    ('s25', 0, 4),
                                                    ('h25', 0, 2),
                                                    ('b25', 0, 1)],
                 alias_names=('v25',), floating_point=True, vector=True),
        Register(name='q26', size=16, subregisters=[('d26', 0, 8),
                                                    ('s26', 0, 4),
                                                    ('h26', 0, 2),
                                                    ('b26', 0, 1)],
                 alias_names=('v26',), floating_point=True, vector=True),
        Register(name='q27', size=16, subregisters=[('d27', 0, 8),
                                                    ('s27', 0, 4),
                                                    ('h27', 0, 2),
                                                    ('b27', 0, 1)],
                 alias_names=('v27',), floating_point=True, vector=True),
        Register(name='q28', size=16, subregisters=[('d28', 0, 8),
                                                    ('s28', 0, 4),
                                                    ('h28', 0, 2),
                                                    ('b28', 0, 1)],
                 alias_names=('v28',), floating_point=True, vector=True),
        Register(name='q29', size=16, subregisters=[('d29', 0, 8),
                                                    ('s29', 0, 4),
                                                    ('h29', 0, 2),
                                                    ('b29', 0, 1)],
                 alias_names=('v29',), floating_point=True, vector=True),
        Register(name='q30', size=16, subregisters=[('d30', 0, 8),
                                                    ('s30', 0, 4),
                                                    ('h30', 0, 2),
                                                    ('b30', 0, 1)],
                 alias_names=('v30',), floating_point=True, vector=True),
        Register(name='q31', size=16, subregisters=[('d31', 0, 8),
                                                    ('s31', 0, 4),
                                                    ('h31', 0, 2),
                                                    ('b31', 0, 1)],
                 alias_names=('v31',), floating_point=True, vector=True),
        Register(name='qcflag', size=16, floating_point=True),
        Register(name='emnote', size=4),
        Register(name='cmstart', size=8),
        Register(name='cmlen', size=8),
        Register(name='nraddr', size=8),
        Register(name='ip_at_syscall', size=8, artificial=True),
        Register(name='fpcr', size=4, floating_point=True, default_value=(initial_sp, True, 'global')),
    ]

    got_section_name = '.got'
    ld_linux_name = 'ld-linux-aarch64.so.1'
    elf_tls = TLSArchInfo(1, 32, [], [0], [], 0, 0)

register_arch([r'.*arm64.*|.*aarch64*'], 64, 'any', ArchAArch64)
